import dataclasses
from typing import List
from typing import Union
from typing import Optional


@dataclasses.dataclass(frozen=True)
class KeyPair:
    kind: str
    name: Union[str, int]

    def __post_init__(self):
        if not isinstance(self.kind, str):
            raise ValueError(
                f'kind must be an instance of str, but received: {type(self.kind)} (value: {self.kind})'
            )

        if not (isinstance(self.name, str) or isinstance(self.name, int)):
            raise ValueError(
                f'name must be an instance of str or int, but received: {type(self.name)} (value: {self.name})'
            )

        if self.kind.find("'") >= 0 or self.kind.find('"') >= 0:
            raise ValueError(f'Invalid kind of "{self.kind}", do not include quotes in kind')

        if isinstance(self.name, str):
            if self.name.find("'") >= 0 or self.name.find('"') >= 0:
                raise ValueError(f'Invalid name of "{self.name}", do not include quotes in name')

    @classmethod
    def build(cls, kind: str, name: Union[str, int], implicit_id_str: bool = True):
        if implicit_id_str:
            if isinstance(name, str) and name.isdecimal():
                name = int(name)

        return cls(kind=kind, name=name)

    @classmethod
    def build_from_dict(cls, doc: dict):
        if 'kind' not in doc:
            raise ValueError(f'dictionary must have "kind" item, but received doc={doc}')
        if 'name' not in doc and 'id' not in doc:
            raise ValueError(f'dictionary must have "name" or "id" item, but received doc={doc}')

        return cls(doc.get('kind'), doc.get('name', doc.get('id')))

    def is_name(self) -> bool:
        return isinstance(self.name, str)

    def is_id(self) -> bool:
        return isinstance(self.name, int)

    def key_pair_literal(self) -> str:
        if self.is_name():
            return f"'{self.kind}', '{self.name}'"
        else:
            return f"'{self.kind}', {self.name}"


class _BaseKey:
    def parent(self):
        raise NotImplementedError()

    def pairs(self) -> List[KeyPair]:
        return []

    def flat_pairs(self) -> List[Union[str, int]]:
        return []

    def kind(self) -> str:
        raise NotImplementedError()

    def name(self) -> Union[str, int]:
        raise NotImplementedError()

    def key_literal(self) -> str:
        raise NotImplementedError()

    def key_path(self) -> str:
        raise NotImplementedError()

    def key_path_urlsafe(self) -> str:
        raise NotImplementedError()


class NoneKey(_BaseKey):
    _root = None

    @classmethod
    def get_instance(cls):
        if cls._root:
            return cls._root
        cls._root = cls()
        return cls._root

    def parent(self):
        return self

    def pairs(self):
        return []

    def flat_pairs(self):
        return []

    def kind(self):
        return None

    def name(self):
        return None

    def __repr__(self):
        return "NoneKey(kind='None', name='none')"

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def key_literal(self):
        return None

    def key_path(self):
        return None

    def key_path_urlsafe(self):
        return None


@dataclasses.dataclass(frozen=True)
class EntityKey(_BaseKey):
    _pairs: List[KeyPair]

    def __post_init__(self):
        if len(self._pairs) == 0:
            raise ValueError('EntityKey#pairs must have at least one element.')

    def parent(self):
        if len(self._pairs) == 1:
            return NoneKey.get_instance()

        return EntityKey(self._pairs[0:-1])

    def pairs(self):
        return self._pairs

    def flat_pairs(self):
        flat_pairs = []
        for pair in self.pairs():
            flat_pairs.append(pair.kind)
            flat_pairs.append(pair.name)

        return flat_pairs

    def kind(self):
        return self._pairs[-1].kind

    def name(self):
        return self._pairs[-1].name

    def key_literal(self) -> str:
        return 'Key({})'.format(', '.join([
            pair.key_pair_literal() for pair in self.pairs()
        ]))

    def key_path(self) -> str:
        """
        Returns a structured key separated by `/`
        The kind and name are separated by `:`
        """
        pairs = []
        for pair in self.pairs():
            pairs.append(f'{pair.kind}:{pair.name}')
        return '/'.join(pairs)

    def key_url(self) -> str:
        """
        Returns a structured key separated by / between kind and name.
        This format is suitable for expressing in a form embedded in a URL.
        """
        return '/'.join([
            f'{pair.kind}/{pair.name}' for pair in self.pairs()
        ])

    def key_path_urlsafe(self) -> str:
        return self.key_path().replace('/', '%2F').replace(':', '%3A')


@dataclasses.dataclass(frozen=True)
class IncompleteKey:
    kind: str
    parent: Optional[EntityKey] = None

    def __post_init__(self):
        if not isinstance(self.kind, str):
            raise ValueError(f'kind must be an instance of str, but received {type(self.kind)} (value: {self.kind})')

        if not (self.parent is None or isinstance(self.parent, EntityKey)):
            raise ValueError(
                f'parent must be an instance of EntityKey, but received {type(self.parent)} (value: {self.parent}'
            )

    def flat_pairs(self):
        flat_pairs = self.parent.flat_pairs() if self.parent else []
        flat_pairs.append(self.kind)

        return flat_pairs

    def key_literal(self) -> str:
        if self.parent:
            return 'IncompleteKey({}, {})'.format(
                ', '.join([pair.key_pair_literal() for pair in self.parent.pairs()]),
                self.kind
            )
        return 'IncompleteKey({})'.format(self.kind)



import os
from logging import getLogger
from typing import Optional

from gumo.core.injector import injector
from gumo.core.domain.configuration import GumoConfiguration
from gumo.core.domain.configuration import GoogleCloudLocation
from gumo.core.domain.configuration import GoogleCloudProjectID
from gumo.core.domain.configuration import ApplicationPlatform
from gumo.core.domain.configuration import ServiceAccountCredentialPath
from gumo.core.exceptions import ConfigurationError

logger = getLogger('gumo.core')


class ConfigurationFactory:
    @classmethod
    def build(
            cls,
            google_cloud_project: Optional[str] = None,
            google_cloud_location: Optional[str] = None,
            service_account_credential_path: Optional[str] = None,
    ) -> GumoConfiguration:

        project_id = GoogleCloudProjectID(
            google_cloud_project if google_cloud_project else os.environ.get('GOOGLE_CLOUD_PROJECT')
        )

        location = GoogleCloudLocation(
            google_cloud_location if google_cloud_location else os.environ.get('GOOGLE_CLOUD_LOCATION')
        )

        is_google_platform = 'GAE_DEPLOYMENT_ID' in os.environ and 'GAE_INSTANCE' in os.environ
        application_platform = ApplicationPlatform.GoogleAppEngine if is_google_platform else ApplicationPlatform.Local

        if application_platform == ApplicationPlatform.Local:
            if 'GOOGLE_APPLICATION_CREDENTIALS' not in os.environ:
                raise ConfigurationError('Envrionment variable "GOOGLE_APPLICATION_CREDENTIALS" is required.')

        credential_path = ServiceAccountCredentialPath(
            value=service_account_credential_path
        )

        return GumoConfiguration(
            google_cloud_project=project_id,
            google_cloud_location=location,
            application_platform=application_platform,
            service_account_credential_config=credential_path.credential_config(),
        )


def configure(
        google_cloud_project: Optional[str] = None,
        google_cloud_location: Optional[str] = None,
        service_account_credential_path: Optional[str] = None,
):
    config = ConfigurationFactory.build(
        google_cloud_project=google_cloud_project,
        google_cloud_location=google_cloud_location,
        service_account_credential_path=service_account_credential_path,
    )
    logger.debug(f'Gumo is configured, config={config}')

    if 'GOOGLE_CLOUD_PROJECT' not in os.environ:
        logger.debug('Environment Variable "GOOGLE_CLOUD_PROJECT" is not configured.')
        project_id = config.google_cloud_project.value
        os.environ['GOOGLE_CLOUD_PROJECT'] = project_id
        logger.debug(f'Environment Variable "GOOGLE_CLOUD_PROJECT" has been updated to {project_id}')

    injector.binder.bind(GumoConfiguration, config)

    return config



from typing import Union
from typing import List

import uuid
import base64
import enum

from gumo.core.domain.entity_key import EntityKey
from gumo.core.domain.entity_key import IncompleteKey
from gumo.core.domain.entity_key import EntityKeyFactory

from gumo.core.injector import injector


class KeyGenerateStyle(enum.Enum):
    STR_SHORT = 'short-str'
    STR_LONG = 'long-str'
    INT = 'int'


class KeyIDAllocator:
    def allocate(self, incomplete_key: IncompleteKey) -> EntityKey:
        raise NotImplementedError()


class EntityKeyGenerator:
    def __init__(self, key_generate_style: KeyGenerateStyle = KeyGenerateStyle.STR_SHORT):
        self._key_generate_style = key_generate_style

        if self._key_generate_style == KeyGenerateStyle.INT:
            self._id_allocator = self._get_id_allocator()

    def _get_id_allocator(self) -> KeyIDAllocator:
        try:
            import gumo.datastore
        except ImportError:
            raise RuntimeError(f'gumo.datastore is not imported. Cloud not use KeyGenerateStyle.INT.')

        return injector.get(KeyIDAllocator)

    def generate(self, incomplete_key: IncompleteKey) -> EntityKey:
        return EntityKeyFactory().build(
            kind=incomplete_key.kind,
            name=self.generate_identifier(incomplete_key=incomplete_key),
            parent=incomplete_key.parent,
        )

    def generate_identifier(self, incomplete_key: IncompleteKey) -> Union[str, int]:
        if self._key_generate_style == KeyGenerateStyle.STR_SHORT:
            return self._generate_str_short()
        elif self._key_generate_style == KeyGenerateStyle.STR_LONG:
            return self._generate_str_long()
        elif self._key_generate_style == KeyGenerateStyle.INT:
            return self._generate_int(incomplete_key=incomplete_key)
        else:
            raise NotImplementedError(f'Not implemented pattern of {self._key_generate_style}')

    def _generate_str_short(self) -> str:
        s = base64.b32encode(uuid.uuid4().bytes).decode('utf-8')
        return s.replace('======', '').lower()

    def _generate_str_long(self) -> str:
        return uuid.uuid4().bytes.decode('utf-8')

    def _generate_int(self, incomplete_key: IncompleteKey) -> int:
        key = self._id_allocator.allocate(incomplete_key=incomplete_key)
        return key.name()



import os
import json

from typing import Optional
from typing import Union
from typing import Tuple

import google.auth.transport.requests
from google.oauth2 import service_account
from google.auth import compute_engine
from google.cloud import storage
from injector import inject

from gumo.core.exceptions import ServiceAccountConfigurationError
from gumo.core.injector import injector
from gumo.core.domain.configuration import GumoConfiguration

DEFAULT_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'

Credentials = Union[service_account.Credentials, compute_engine.Credentials]
IDTokenCredentials = Union[service_account.IDTokenCredentials, compute_engine.IDTokenCredentials]
Request = google.auth.transport.Request


class GoogleOAuthCredentialManager:
    @inject
    def __init__(
            self,
            gumo_configuration: GumoConfiguration,
    ):
        self._gumo_configuration = gumo_configuration
        self._credential_config = self._gumo_configuration.service_account_credential_config

    def build_credential(self) -> Credentials:
        _credentials = None

        try:
            if self._gumo_configuration.is_google_app_engine:
                _credentials = compute_engine.Credentials()
            elif 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ:
                _credentials = service_account.Credentials.from_service_account_info(
                    info=self._get_content_from_local()
                )
            elif self._credential_config.enabled:
                # will be deprecated.
                _credentials = service_account.Credentials.from_service_account_info(
                    info=self._get_content_from_storage()
                )
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

        if _credentials is None:
            raise ServiceAccountConfigurationError(f'ServiceAccount Credential Config disabled.')

        return _credentials

    def build_request(self) -> Request:
        return google.auth.transport.requests.Request()

    def build_id_token_credential(
            self,
            target_audience: str,
            with_refresh: bool = True,
            token_uri: Optional[str] = None,
    ) -> Tuple[IDTokenCredentials, Request]:
        request = self.build_request()
        _id_token_credential = None

        try:
            if self._gumo_configuration.is_google_app_engine:
                _id_token_credential = compute_engine.IDTokenCredentials(
                    request=request,
                    target_audience=target_audience,
                )
                if with_refresh:
                    _id_token_credential.refresh(request=request)
            elif 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ:
                _credential = self.build_credential()
                _id_token_credential = service_account.IDTokenCredentials(
                    signer=_credential.signer,
                    service_account_email=_credential.service_account_email,
                    token_uri=token_uri if token_uri else DEFAULT_TOKEN_URI,
                    target_audience=target_audience,
                )
                if with_refresh:
                    _id_token_credential.refresh(request=request)
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

        if _id_token_credential is None:
            raise ServiceAccountConfigurationError(f'ServiceAccount Credential Config disabled.')

        return (_id_token_credential, request)

    def _get_content_from_storage(self):
        storage_client = storage.Client()
        bucket = storage_client.get_bucket(bucket_name=self._credential_config.bucket_name)
        blob = bucket.blob(blob_name=self._credential_config.blob_path)
        content = blob.download_as_string(client=storage_client)

        return json.loads(content)

    def _get_content_from_local(self):
        credential_path = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')

        if not os.path.exists(credential_path):
            raise ServiceAccountConfigurationError(f'GOOGLE_APPLICATION_CREDENTIALS={credential_path} is not found.')

        with open(credential_path, 'r') as f:
            content = f.read()

        return json.loads(content)


def get_google_oauth_credential() -> Credentials:
    return injector.get(GoogleOAuthCredentialManager).build_credential()


def get_google_id_token_credential(
        target_audience: str,
        with_refresh: bool = True,
        token_uri: Optional[str] = None,
) -> Tuple[IDTokenCredentials, Request]:
    return injector.get(GoogleOAuthCredentialManager).build_id_token_credential(
        target_audience=target_audience,
        with_refresh=with_refresh,
        token_uri=token_uri,
    )



from typing import List
from typing import Union
from typing import Optional

import copy
import base64
import uuid

from gumo.core.domain.entity_key.key import EntityKey
from gumo.core.domain.entity_key.key import KeyPair
from gumo.core.domain.entity_key.key import NoneKey
from gumo.core.domain.entity_key.key import IncompleteKey


class EntityKeyFactory:
    def __init__(self):
        pass

    def build_from_pairs(self, pairs: List[Union[tuple, dict]]) -> EntityKey:
        _pairs = []
        for pair in pairs:
            if isinstance(pair, dict):
                key_pair = KeyPair.build_from_dict(doc=pair)
            elif isinstance(pair, tuple) or isinstance(pair, list):
                key_pair = KeyPair(pair[0], pair[1])
            else:
                raise ValueError(f'Unknown object type: {type(pair)}, got: {pair}')
            _pairs.append(key_pair)

        return EntityKey(_pairs)

    def build(self, kind: str, name: Union[str, int], parent: Optional[EntityKey] = None) -> EntityKey:
        if parent:
            pairs = copy.deepcopy(parent.pairs())
        else:
            pairs = []
        pairs.append(KeyPair(kind=kind, name=name))

        return EntityKey(pairs)

    def build_incomplete_key(self, kind: str, parent: Optional[EntityKey] = None) -> IncompleteKey:
        return IncompleteKey(kind=kind, parent=parent)

    def build_for_new(self, kind: str, parent: Optional[EntityKey] = None) -> EntityKey:
        """(deprecated) Please use gumo.core.application.EntityKeyGenerator."""
        name = self._generate_new_uuid()
        return self.build(kind=kind, name=name, parent=parent)

    def _generate_new_uuid(self) -> str:
        s = base64.b32encode(uuid.uuid4().bytes).decode('utf-8')
        return s.replace('======', '').lower()

    def build_from_key_path(self, key_path: str) -> EntityKey:
        pairs = []
        for pair in key_path.replace('%2F', '/').replace('%3A', ':').split('/'):
            kind, name = pair.split(':')
            pairs.append(KeyPair.build(kind=kind, name=name, implicit_id_str=True))

        return EntityKey(pairs)

    @staticmethod
    def _split_list(l, n):
        """
        list divided by n items.
        :param l: list
        :param n: item size of sub-list
        :return:
        """
        for idx in range(0, len(l), n):
            yield l[idx:idx + n]

    def build_from_key_url(self, key_url: str) -> EntityKey:
        pairs = []

        key_elements = key_url.replace('%2F', '/').split('/')
        key_pairs = list(self._split_list(key_elements, 2))

        for pair in key_pairs:
            kind, name = pair
            pairs.append(KeyPair.build(kind=kind, name=name, implicit_id_str=True))

        return EntityKey(pairs)
